
// general settings

package HOTCOInnoSimV15J;

import java.text.SimpleDateFormat;
import java.util.Date;

import java.util.Scanner;
import java.util.Locale;

import java.util.LinkedList;
import java.util.HashSet;

import java.util.Random;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.BufferedWriter;
import java.io.FileWriter;

import static java.lang.Math.abs;
import static java.lang.Math.min;
import static java.lang.Math.max;
import static java.lang.Math.pow;
import static java.lang.Math.sqrt;

import static java.lang.Double.NaN;

import static java.util.Arrays.sort;

public class HOTCOInnoSimV15
{
    public String file;
    
    // do not change the values here
    // those are defaults
    // use the main function instead
    
    public String _network_type_ = "social-radius";
    public String _social_mechanisms_ = "means-ends & contagion";
    
    public double _goal_excit_ = 0.04;
    public double _excit_weight_ = 0.04;
    public double _inhib_weight_ = 0.04;
    public double _max_activation_ = + 0.99;
    public double _min_activation_ = - 0.99;
    public double _decay_amount_ = 0.05;
    public double _output_threshold_ = - 0.99;
    public double _asymptote_ = 0.0001;
    public double _valence_weight_ = 0.05;
    
    public double _random_partners_ = 0.05;
    public double _ex_radius_ = 2.5;
    
    public double _link_chance_ = 0.5;
    
    public double _activation_default_value_ = 0.01;
    public double _new_activation_default_value_ = NaN;
    public double _valence_default_value_ = 0.01;
    public double _new_valence_default_value_ = 0.01;
    
    public double _valence_threshold_ = 0.1;
    
    public double _activated_ = 0.2;
    
    
    public HOTCOInnoSimV15 ( String file )
    {
        this.file = file;
		this.random = new Random ();
    }
	
    public HOTCOInnoSimV15 ( String file, long seed )
    {
        this.file = file;
		this.random = new Random ( seed );
    }
    
    public final Random random;
    
    public class Unit
    {
        public String name;
        
        public Unit ( String name )
        {
            this.name = name;
        }
        
        public double activation;
        public double activation_new;
        
        public double valence;
        public double valence_new;
        
        public LinkedList < LinkTo > u_links = new LinkedList < LinkTo > ();
        public LinkedList < LinkTo > v_links = new LinkedList < LinkTo > ();
    }
    
    public class LinkTo
    {
        public Unit unit;
        public Link link;
        
        public LinkTo ( Unit unit, Link link )
        {
            this.unit = unit;
            this.link = link;
        }
    }
    
    public class Link
    {
        public Unit unit_a;
        public Unit unit_b;
        
		public double weight_new;
		public double weight_old;
		
		public double weight_neg;
		public double weight_pos;
		
        public Link ( Unit unit_a, Unit unit_b, double weight, double weight_neg, double weight_pos )
        {
            this.unit_a = unit_a;
            this.unit_b = unit_b;
			
			this.weight_new = weight;
			this.weight_old = weight;
			
			this.weight_neg = weight_neg;
			this.weight_pos = weight_pos;
        }
		
		public double eval ()
		{
			if ( weight_old < 0.0 ) return weight_old * weight_neg;
			if ( weight_old > 0.0 ) return weight_old * weight_pos;
			
			return 0.0;
		}
    }
    
    public class Graph
    {
        public Graph ()
        { }
        
        public Unit spunit;
        public Unit svunit;
        
		public Link [] links = new Link [ 8 + 8 + 5 + 5 + 5 * 8 ];
		
        public Unit [] issues = new Unit [ 5 ];
        
        public Unit [] goals = new Unit [ 8 ];
        
		public Link [] ispalinks = new Link [ 5 ];
        public Link [] gspalinks = new Link [ 8 ];
        
        public Link [] ispvlinks = new Link [ 5 ];
        public Link [] gspvlinks = new Link [ 8 ];
        
        public Link [][] facilitations = new Link [ 5 ] [ 8 ];
        
        public LinkedList < Unit > all_units = new LinkedList < Unit > ();
    }
    
    public class Person
    {
        public String name;
        
        public int index;
        
        public Person ( String name )
        {
            this.name = name;
        }
        
        public Graph graph;
        
        // demogaphics
        
        public double [] demographics;
        
        public int age;
        public int gender;
        public int income;
        public int education;
        public double equipment;
        public double moderness;
        public int cluster;
        public double position_x;
        public double position_y;
        public double radius;
        
        public double x;
        public double y;
        
		public double media = 1.0; 
		
        // 
        
        public int step;
        
        // 
        
        public LinkedList < Person > temp = new LinkedList < Person > ();
        
        public Partner [] partners;
        
        public LinkedList < ConnectionTo > connections = new LinkedList < ConnectionTo > ();
        
        public double node_clustering_coefficient;
    }
    
    public LinkedList < Connection > connections = new LinkedList < Connection > ();
    
    public class Partner
    {
        public Person person;
        
        public Partner ( Person person )
        {
            this.person = person;
        }
        
        public double distance, socialweight;
    }
    
    public class Connection
    {
        public Person person_a;
        public Person person_b;
        
        public Connection ( Person person_a, Person person_b )
        {
            this.person_a = person_a;
            this.person_b = person_b;
        }
    }
    
    public class ConnectionTo
    {
        public Person person;
        
        public ConnectionTo ( Person person )
        {
            this.person = person;
        }
    }
    
    public String [] issues
        = new String [] { "Car", "E-Car", "Car-Sharing", "Public-Transport", "Bicycle" };
        
    public String [] goals
        = new String [] { "Security", "Comfort", "Ecology", "Experience", "Image", "Cost", "Stress", "Independence" };
    
    public Person [] persons;
    
    public Graph graph ( Scanner scanner )
    {
               
        double [] input_a = new double [ 8 ];
        double [] input_b = new double [ 8 ];
        
        double [][] input_f = new double [ 5 ] [ 8 ];
        
        for ( int index = 0; index < 8; index ++ )
        {
            input_a [ index ] = scanner.nextDouble ();
        }
        
        for ( int index = 0; index < 8; index ++ )
        {
            input_b [ index ] = scanner.nextDouble ();
        }
        
        for ( int index_a = 0; index_a < 5; index_a ++ )
        {
            for ( int index_b = 0; index_b < 8; index_b ++ )
            {
                input_f [ index_a ] [ index_b ] = scanner.nextDouble ();
            }
        }
        
        // 
        
        Graph graph = new Graph ();
        
        // 
        
        graph.spunit = new Unit ( "Special Unit" );
        graph.svunit = new Unit ( "Special Valence Unit" );
        
        for ( int index = 0; index < 5; index ++ )
        {
            graph.issues [ index ] = new Unit ( issues [ index ] );
        }
        
        for ( int index = 0; index < 8; index ++ )
        {
            graph.goals [ index ] = new Unit ( goals [ index ] );
        }
        
        for ( int index = 0; index < 5; index ++ )
        {
            graph.ispalinks [ index ] = new Link ( graph.spunit, graph.goals [ index ], 0.0, 1.0, 1.0 );
        }
        
        for ( int index = 0; index < 8; index ++ )
        {
            graph.gspalinks [ index ] = new Link ( graph.spunit, graph.goals [ index ], input_a [ index ], _goal_excit_, _goal_excit_ );
        }
        
        for ( int index = 0; index < 5; index ++ )
        {
            graph.ispvlinks [ index ] = new Link ( graph.svunit, graph.issues [ index ], 0.0, 1.0, 1.0 );
        }
        
        for ( int index = 0; index < 8; index ++ )
        {
            graph.gspvlinks [ index ] = new Link ( graph.svunit, graph.goals [ index ], input_b [ index ], _valence_weight_, _valence_weight_ );
        }
        
        for ( int index_a = 0; index_a < 5; index_a ++ )
        {
            for ( int index_b = 0; index_b < 8; index_b ++ )
            {
                graph.facilitations [ index_a ] [ index_b ] = new Link ( graph.issues [ index_a ], graph.goals [ index_b ], input_f [ index_a ] [ index_b ], _inhib_weight_, _excit_weight_ );
            }
        }
        
        // 
        
		int p = 0;
		
        for ( Link link : graph.gspalinks )
        {
			graph.links [ p ++ ] = link;
			
            Unit unit_a = link.unit_a;
            Unit unit_b = link.unit_b;
            
            unit_a.u_links.add ( new LinkTo ( unit_b, link ) );
            unit_b.u_links.add ( new LinkTo ( unit_a, link ) );
        }
        
        for ( Link link : graph.ispalinks )
        {
			graph.links [ p ++ ] = link;
			
            Unit unit_a = link.unit_a;
            Unit unit_b = link.unit_b;
            
            unit_a.u_links.add ( new LinkTo ( unit_b, link ) );
            unit_b.u_links.add ( new LinkTo ( unit_a, link ) );
        }
        
        for ( Link link : graph.ispvlinks )
        {
			graph.links [ p ++ ] = link;
			
            Unit unit_a = link.unit_a;
            Unit unit_b = link.unit_b;
            
            unit_a.v_links.add ( new LinkTo ( unit_b, link ) );
            unit_b.v_links.add ( new LinkTo ( unit_a, link ) );
        }
        
        for ( Link link : graph.gspvlinks )
        {
			graph.links [ p ++ ] = link;
			
            Unit unit_a = link.unit_a;
            Unit unit_b = link.unit_b;
            
            unit_a.v_links.add ( new LinkTo ( unit_b, link ) );
            unit_b.v_links.add ( new LinkTo ( unit_a, link ) );
        }
        
        for ( Link [] links : graph.facilitations )
        {
            for ( Link link : links )
            {
				graph.links [ p ++ ] = link;
				
                Unit unit_a = link.unit_a;
                Unit unit_b = link.unit_b;
                
                unit_a.u_links.add ( new LinkTo ( unit_b, link ) );
                unit_b.u_links.add ( new LinkTo ( unit_a, link ) );
            }
        }
        
        for ( Unit unit : graph.goals )
        {
            graph.all_units.add ( unit );
        }
        
        for ( Unit unit : graph.issues )
        {
            graph.all_units.add ( unit );
        }
        
        // 
        
        return graph;
    }
	
	public static enum MEDIA { RELATIVE, ABSOLUTE } 
	
	public static class MediaAgent2
	{
		public final String source;
		public final MEDIA media;
		public final double factor;
		
		public final double [] input_a = new double [ 8 ];
		public final double [] input_b = new double [ 8 ];
		
		public final double [] input_c = new double [ 5 ];
		public final double [] input_d = new double [ 5 ];
		
		public final double [][] input_f = new double [ 5 ] [ 8 ];
		
		
		public MediaAgent2 ( String source, MEDIA media, double factor )
		{
			this.source = source;
			this.media = media;
			this.factor = factor;
			
			try
            {
                Scanner scanner = new Scanner( new File ( source ), "UTF-8" );
                
                scanner.nextLine();
                
                try
                {
                    scanner.useLocale ( Locale.ENGLISH );
                    
                    // goal -> sp unit
					// goal -> sv unit
					// facilitations
					// goal -> valence
					// issues -> valence
					
					for ( int index = 0; index < 8; index ++ )
					{
						input_a [ index ] = scanner.nextDouble ();
					}
					
					for ( int index = 0; index < 8; index ++ )
					{
						input_b [ index ] = scanner.nextDouble ();
					}
					
					for ( int index = 0; index < 5; index ++ )
					{
						input_c [ index ] = scanner.nextDouble ();
					}
					
					for ( int index = 0; index < 5; index ++ )
					{
						input_d [ index ] = scanner.nextDouble ();
					}
					
					for ( int index_a = 0; index_a < 5; index_a ++ )
					{
						for ( int index_b = 0; index_b < 8; index_b ++ )
						{
							input_f [ index_a ] [ index_b ] = scanner.nextDouble ();
						}
					}
                }
                finally
                {
                    scanner.close ();
                }
            }
            catch ( FileNotFoundException exception )
            {
                throw new UnsupportedOperationException ( exception );
            }
		}
	}
	
    public class MediaAgent1
    {
        public final Graph graph;
        
        public final String source;
        
        public MediaAgent1 ( String source )
        {
            this.source = source;
            
            try
            {
                Scanner scanner = new Scanner( new File ( source ), "UTF-8" );
                
                scanner.nextLine();
                
                try
                {
                    scanner.useLocale ( Locale.ENGLISH );
                    
                    graph = graph ( scanner );
                    
                    for ( int index = 0; index < 5; index ++ )
                    {
                        Unit issue = graph.issues [ index ];
                        
                        issue.valence = scanner.nextDouble ();
                    }
                }
                finally
                {
                    scanner.close ();
                }
            }
            catch ( FileNotFoundException exception )
            {
                throw new UnsupportedOperationException ( exception );
            }
        }
    }
    
    public void load ( String source )
    {
        LinkedList < Person > persons = new LinkedList < Person > ();
        
        try
        {
            Scanner scanner = new Scanner( new File ( source ), "UTF-8" );
            
            scanner.nextLine();
            
            try
            {
                scanner.useLocale ( Locale.ENGLISH );
                
                while ( scanner.hasNext () )
                {
                    Person person = new Person ( scanner.next () );
                    
                    person.graph = graph ( scanner );
                    
                    // 
                    
                    person.age            = scanner.nextInt();
                    person.gender         = scanner.nextInt();
                    person.income         = scanner.nextInt();
                    person.education      = scanner.nextInt();
                    person.equipment      = scanner.nextDouble();
                    person.moderness      = scanner.nextDouble();
                    person.cluster        = scanner.nextInt();
                    person.position_x     = scanner.nextDouble();
                    person.position_y     = scanner.nextDouble();
                    person.radius         = scanner.nextDouble();
                    
					person.media          = scanner.nextDouble();
					
                    // 
                    
                    persons.add ( person );
                }
            }
            finally
            {
                scanner.close ();
            }
        }
        catch ( FileNotFoundException exception )
        {
            throw new UnsupportedOperationException ( exception );
        }
        
        // 
        
        this.persons = new Person [ persons.size () ];
        
        int n = 0;
        
        for ( Person person : persons ) this.persons [ n ++ ] = person;
        
        for ( int index = 0; index < n; index ++ )
        {
            Person person = this.persons [ index ];
            
            person.index = index;
        }
    }
    
    public void setup ( String source )
    {
        System.out.println ( String.format ( "setup ( \"%s\" )", source ) );
        
        load ( source );
        
        for ( Person person : persons )
        {
            person.x = person.position_x  + random_double ();
            person.y = person.position_y  + random_double ();
        }
        
		int np = persons.length;
		
		history = new int [ np ] [ np ];
		
        // initialize-hot-persons > hot-person-coh
        
        for ( Person person : persons )
        {
            hot_person_coh ( person );
        }
        
        // 
        
        for ( Person person : persons )
        {
            person.demographics = new double [] {
                person.age,
                person.gender,
                person.income,
                person.education,
                person.equipment,
                person.moderness };
        }
        
        // 
        
        for ( Person person : persons )
        {
            double radius = person.radius * _ex_radius_;
            
            for ( Person partner : persons )
            {
                if ( ( person != partner ) && ( distance ( person, partner ) < radius )  )
                {
                    person.temp.add ( partner );
                }
            }
        }
        
        for ( Person person_a : persons )
        {
            int n = round ( persons.length * _random_partners_ );
            
            for ( int index = 0; index < n; index ++ )
            {
                Person person_b = persons [ random_int ( persons.length ) ];
                
                if ( person_a.temp.contains ( person_b ) )
                { }
                else
                {
                    person_a.temp.add ( person_b );
                    person_b.temp.add ( person_a );
                }
            }
        }
        
        for ( Person person : persons )
        {
            int sup = person.temp.size ();
            
            person.partners = new Partner [ sup ];
            
            for ( Person partner : person.temp )
            {
                int n = random_int ( sup );
                
                while ( person.partners [ n % sup ] != null ) n ++;
                
                person.partners [ n % sup ] = new Partner ( partner );
            }
        }
        
        // 
        
        double neg = Double.NEGATIVE_INFINITY;
        
        _max_d_ = new double [] { neg, neg, neg, neg, neg, neg };
        
        for ( Person person : persons )
        {
            for ( Partner partner : person.partners )
            {
                for ( int index = 0; index < 6; index ++ )
                {
                    _max_d_ [ index ] = max ( _max_d_ [ index ], person.demographics [ index ] - partner.person.demographics [ index ] );
                }
            }
        }
        
        // 
        
        for ( Person person : persons )
        {
			double max_distance = 0.0;
			
            for ( Partner partner : person.partners )
            {
                double distance = 0.0;
                
                for ( int index = 0; index < 6; index ++ )
                {
                    distance += pow ( ( person.demographics [ index ] - partner.person.demographics [ index ] ) / _max_d_ [ index ], 2.0 );
                }
                
                distance = sqrt ( distance );
                
                partner.distance = distance;
                
                max_distance = max ( max_distance, distance );
            }
            
            for ( Partner partner : person.partners )
            {
                partner.socialweight = max( 0.01, 1.0 - partner.distance / max_distance );
            }
        }
        
        // 
        
        if ( false )
        { }
        else if ( _network_type_.equals ( "all-connected" ) )
        {
            setup_links_complete ();
        }
        else if ( _network_type_.equals ( "social-radius" ) )
        {
            setup_links_social_radius ();
        }
        else if ( _network_type_.equals ( "scale-free" ) )
        {
            setup_links_barabasi_albert ();
        }
        else if ( _network_type_.equals ( "random" ) )
        {
            setup_links_erdos_renyi ();
        }
        else
        {
            throw new UnsupportedOperationException ( String.format ( "network-type = \"%s\"", _network_type_ ) );
        }
        
        // 
        
        network_calc ();
        
        // write network stats
		
		try
        {
            BufferedWriter io = new BufferedWriter ( new FileWriter ( file, true ) );
            
            try
            {
				io.write ( String.format ( Locale.ENGLISH, "#%32s\t%8.4f\r\n", "_MinLinks_", _MinLinks_ ) );
				io.write ( String.format ( Locale.ENGLISH, "#%32s\t%8.4f\r\n", "_MaxLinks_", _MaxLinks_ ) );
				io.write ( String.format ( Locale.ENGLISH, "#%32s\t%8.4f\r\n", "_MeanLinks_", _MeanLinks_ ) );
				io.write ( String.format ( Locale.ENGLISH, "#%32s\t%8.4f\r\n", "_SDLinks_", _SDLinks_ ) );
				io.write ( String.format ( Locale.ENGLISH, "#%32s\t%8.4f\r\n", "_MedianLinks_", _MedianLinks_ ) );
				
                io.write ( "#\n" );
                
                io.write ( String.format ( Locale.ENGLISH, "#%32s\t%8.4f\r\n", "ClusteringCoefficient", _clustering_coefficient_ ) );
                io.write ( String.format ( Locale.ENGLISH, "#%32s\t%8.4f\r\n", "AveragePathLength", _average_path_length_ ) );
                
				io.write ( "#\n" );
				
				io.write ( String.format ( Locale.ENGLISH, "#%16s\t%16s\t%16s\n", "name", "partners", "connections" ) );
				
				for ( Person person : persons )
				{
					io.write ( String.format ( Locale.ENGLISH, "#%16s\t%16d\t%16d\n", person.name, person.partners.length, person.connections.size () ) );
				}
				
				io.write ( "#\n" );
            }
            finally
            {
                io.close ();
            }
        }
        catch ( Exception exception )
        {
            throw new UnsupportedOperationException ( exception );
        }
		
		// 
        
        stats_header ();
        
        stats ();
    }
    
    // 
    
    public boolean create_connection ( Person person_a, Person person_b )
    {
        if ( person_a == person_b ) return false;
        
		for ( ConnectionTo connection : person_a.connections )
		{
			if ( connection.person != person_b )
			{ }
			else
			{
				return false;
			}
		}
		
        connections.add ( new Connection ( person_a, person_b ) );
        
        person_a.connections.add ( new ConnectionTo ( person_b ) );
        person_b.connections.add ( new ConnectionTo ( person_a ) );
		
		return true;
    }
    
    public void setup_links_complete ()
    {
        for ( Person a : persons )
        {
            for ( Person b : persons )
            {
                create_connection ( a, b );
            }
        }
    }
    
    public void setup_links_social_radius ()
    {
        for ( Person person : persons )
        {
            for ( Partner partner : person.partners )
            {
                if ( partner.socialweight > random_double () )
                {
                    create_connection ( person, partner.person );
                }
            }
        }
    }
    
    public void setup_links_barabasi_albert ()
    {
        int n = persons.length;
        int first = 1;
        int num_links = 0;
        int choice = 0;
        int [] destinations = new int [ n ];
        
        for ( int index = 0; index < n; index ++ )
        {
            destinations [ index ] = ( - 1 );
        }
        
        create_connection ( persons [ first ], persons [ choice ] );
        
        destinations [ num_links ] = choice;
        
        while ( num_links < ( n - 2 ) )
        {
            choice = random_int ( 2 * ( num_links + 1 ) ) - num_links;
            
            if ( choice < 0 )
            {
                choice = destinations [ - choice ];
            }
            
            ++ num_links;
            
            ++ first;
            
            create_connection ( persons [ first ], persons [ choice ] );
            
            destinations [ num_links ] = choice;
        }
    }
    
    public void setup_links_erdos_renyi ()
    {
        int n = persons.length;
        
        int m = ( ( int ) ( ( ( double ) n * ( n - 1 ) ) * _link_chance_ ) );
        
        boolean [][] map = new boolean [ n ] [ n ];
        
        while ( m > 0 )
        {
            int a = random_int ( n );
            int b = random_int ( n );
            
            if ( ( a > b ) && ! map [ a ] [ b ] )
            {
                m = m - 1;
                
                map [ a ] [ b ] = true;
                
                create_connection ( persons [ a ], persons [ b ] );
            }
        }
    }
    
    public double [] _max_d_;
    
    public double distance ( Person a, Person b )
    {
        return sqrt ( pow ( a.x - b.x, 2.0 ) + pow ( a.y - b.y, 2.0 ) );
    }
    
    // 
    
    public double _average_path_length_;
    
    public void network_calc ()
    {
        link_stats ();
        
        boolean inf = false;
        
        int [][] distances = find_path_lengths ();
        
        double sum = 0.0;
        
        int n = persons.length;
        
        for ( int  a = 0; a < n; a ++ )
        {
            for ( int b = 0; b < n; b ++ )
            {
                int d = distances [ a ] [ b ];
                
                if ( d < 0 )
                {
                    inf = true;
                }
                else
                {
                    sum += d;
                }
            }
        }
        
        if ( inf )
        {
            _average_path_length_ = Double.POSITIVE_INFINITY;
        }
        else
        {
            _average_path_length_ = ( sum / n ) / ( n - 1 );
        }
        
        find_clustering_coefficient ();
    }
    
    public int [] connectors ( Person person )
    {
        int n = person.connections.size ();
        
        int [] values = new int [ n ];
        
        int index = 0;
        
        for ( ConnectionTo connection : person.connections )
        {
            values [ index ++ ] = connection.person.index;
        }
        
        return values;
    }
    
    public void find_clustering_coefficient ()
    {
        int n = persons.length;
        
        int [][] values = new int [ n ] [];
        
        int [] map = new int [ n ];
        
        for ( int index = 0; index < n; index ++ )
        {
            values [ index ] = connectors ( persons [ index ] );
        }
        
        for ( int index = 0; index < n; index ++ )
        {
            sort ( values [ index ] );
        }
        
        for ( Connection connection : connections )
        {
            int a = connection.person_a.index;
            int b = connection.person_b.index;
            
            int na = values [ a ].length;
            int nb = values [ b ].length;
            
            int ia = 0;
            int ib = 0;
            
            while ( ( ia < na ) && ( ib < nb ) )
            {
                int va = values [ a ] [ ia ];
                int vb = values [ b ] [ ib ];
                
                if ( va != vb )
                {
                    if ( va < vb ) ia ++;
                    if ( va > vb ) ib ++;
                }
                else
                {
                    ++ map [ va ];
                    
                    ++ ia;
                    ++ ib;
                }
            }
        }
        
        double total = 0.0;
        double count = 0.0;
        
        for ( int index = 0; index < n; index ++ )
        {
            Person person = persons [ index ];
            
            double coef = ( 2.0 * ( map [ index ] ) ) / ( n * ( n - 1.0 ) );
            
            total = total + coef;
            count = count + 1.0;
            
            person.node_clustering_coefficient = coef;
        }
        
        _clustering_coefficient_ = count > 0.0 ? total / count : 0.0;
    }
    
    public double _clustering_coefficient_;
    
    public int [][] find_path_lengths ()
    {
        int n = persons.length;
        
        int [][] distances = new int [ n ] [ n ];
        
        for ( int a = 0; a < n; a ++ )
        {
            for ( int b = 0; b < n; b ++ )
            {
                if ( a != b )
                {
                    distances [ a ] [ b ] = - 1;
                }
                else
                {
                    distances [ a ] [ b ] = 0;
                }
            }
        }
        
        for ( Connection connection : connections )
        {
            Person person_a = connection.person_a;
            Person person_b = connection.person_b;
            
            distances [ person_a.index ] [ person_b.index ] = 1;
            distances [ person_b.index ] [ person_a.index ] = 1;
        }
        
        for ( int k = 0; k < n; k ++ )
        {
            for ( int a = 0; a < n; a ++ )
            {
                for ( int b = 0; b < n; b ++ )
                {
                    int d = distances [ a ] [ k ] + distances [ k ] [ b ];
                    
                    if ( d < distances [ a ] [ b ] )
                    {
                        distances [ a ] [ b ] = d;
                    }
                }
            }
        }
        
        return distances;
    }
    
    public double _MinLinks_;
    public double _MaxLinks_;
    public double _MeanLinks_;
    public double _SDLinks_;
    public double _MedianLinks_;
    
    public void link_stats ()
    {
        int n = persons.length;
        
        double [] values = new double [ n ];
        
        for ( int index = 0; index < persons.length; index ++ )
        {
        
            values [ index ] = persons [ index ] .connections.size ();
        }
        
        _MinLinks_ = Double.POSITIVE_INFINITY;
        _MaxLinks_ = Double.NEGATIVE_INFINITY;
        
        double s = 0.0;
        double d = 0.0;
        
        for ( int index = 0; index < values.length; index ++ )
        {
            double value = values [ index ];
            
            s += value;
            d += value * value;
            
            _MinLinks_ = min ( _MinLinks_, value );
            _MaxLinks_ = max ( _MaxLinks_, value );
        }
        
        _MeanLinks_ = s / n;
        _SDLinks_ = d / n - ( s / n ) * ( s / n );
        
        sort ( values );
        
        if ( n > 0 )
        {
            int m = n / 2;
            int r = n % 2;
            
            if ( r != 0 )
            {
                _MedianLinks_ = ( values [ m ] );
            }
            else
            {
                _MedianLinks_ = ( values [ m ] + values [ m - 1 ] ) / 2.0;
            }
        }
        else
        {
            _MedianLinks_ = NaN;
        }
    }
    
    // 
    
    public void clear ( Person person )
    {
        for ( Unit unit : person.graph.issues )
        {
            unit.activation = _activation_default_value_;
            unit.activation_new = _new_activation_default_value_;
            unit.valence = _valence_default_value_;
            unit.valence_new = _new_valence_default_value_;
        }
        for ( Unit unit : person.graph.goals )
        {
            unit.activation = _activation_default_value_;
            unit.activation_new = _new_activation_default_value_;
            unit.valence = _valence_default_value_;
            unit.valence_new = _new_valence_default_value_;
        }
        
        person.graph.svunit.activation = 1.0;
        person.graph.svunit.activation_new = _new_activation_default_value_;
        person.graph.svunit.valence = 1.0;
        person.graph.svunit.valence_new = _new_valence_default_value_;
        
        person.graph.spunit.activation = 1.0;
        person.graph.spunit.activation_new = _new_activation_default_value_;
        person.graph.spunit.valence = _valence_default_value_;
        person.graph.spunit.valence_new = _new_valence_default_value_;
    }
    
    public void hot_person_coh ( Person person )
    {
        clear ( person );
        
        run_hyp_net ( person );
    }
    
    public void run_hyp_net ( Person person )
    {
        int steps = 300, step = 0;
        
        boolean settled = false;
        
        while ( ( ! settled ) && ( ( step ++ ) < steps ) )
        {
            settled = true;
            
            update_valences ( person );
            
            for ( Unit unit : person.graph.all_units )
            {
                if ( abs ( unit.activation - unit.activation_new ) > _asymptote_ )
                {
                    settled = false;
                }
                
                unit.activation = unit.activation_new;
            }
            
            for ( Unit unit : person.graph.all_units )
            {
                unit.valence = unit.valence_new;
            }
        }
        
        person.step = step;
    }
    
    public void update_valences ( Person person )
    {
        for ( Unit unit : person.graph.goals )
        {
            update_valence ( person, unit );
        }
        
        for ( Unit unit : person.graph.issues )
        {
            update_valence ( person, unit );
        }
        
        for ( Unit unit : person.graph.goals )
        {
            update_eval_activn ( person, unit );
        }
        
        for ( Unit unit : person.graph.issues )
        {
            update_eval_activn ( person, unit );
        }
    }
    
    public void update_valence ( Person person, Unit unit )
    {
        double net_valence_value = valence_input ( unit );
        
        double tmp = unit.valence * ( 1.0 - _decay_amount_ );
        
        if ( net_valence_value > 0.0 )
        {
            tmp += net_valence_value * ( _max_activation_ - unit.valence );
        }
        
        if ( net_valence_value < 0.0 )
        {
            tmp += net_valence_value * ( unit.valence - _min_activation_ );
        }
        
        unit.valence_new = clamp( tmp );
    }
    
    public void update_eval_activn ( Person person, Unit unit )
    {
        double net_input_value = net_eval_input ( unit );
        
        double tmp = unit.activation * ( 1.0 - _decay_amount_ );
        
        if ( net_input_value > 0.0 )
        {
            tmp += net_input_value * ( _max_activation_ - unit.activation );
        }
        
        if ( net_input_value < 0.0 )
        {
            tmp += net_input_value * ( unit.activation - _min_activation_ );
        }
        
        unit.activation_new = clamp ( tmp );
    }
    
	public double clamp ( double value )
	{
		return min( max( value, _min_activation_ ), _max_activation_ );
	}
	
  
    public double valence_input ( Unit unit )
    {
        double result = 0.0;
        
        for ( LinkTo link : unit.u_links )
        {
			double weight = link.link.eval ();
			
            result += weight * link.unit.activation * link.unit.valence;
        }
        
        for ( LinkTo link : unit.v_links )
        {
			double weight = link.link.eval ();
			
            result += weight * link.unit.activation * link.unit.valence;
        }
        
        return result;
    }
    
    public double net_eval_input ( Unit unit )
    {
        return net_input ( unit ) + valence_input ( unit );
    }
    
    public double net_input ( Unit unit )
    {
        double result = 0.0;
        
        for ( LinkTo link : unit.u_links )
        {
			double weight = link.link.eval ();
			
            result += weight * max ( _output_threshold_, link.unit.activation );
        }
        
        return result;
    }
    
    public int num = 0;
    
	
	public static enum HISTORY { NONE, CYCLE, TOTAL }
	
	public HISTORY _history_ = HISTORY.NONE;
	
	public int [][] history;
	
    public Person one_of ( Person person )
    {
		int u = person.index;
		
        int n = person.connections.size ();
        
        if ( n > 0 )
        {
			int r = random_int ( n );
			
			for ( int i = 0; i < n; i ++ )
			{
				ConnectionTo connection = person.connections.get ( ( i + r ) % n );
				
				Person partner = connection.person;
				
				if ( partner != person )
				{
					switch ( _history_ )
					{
						case NONE :
							
							return partner;
							
						case CYCLE :
						case TOTAL :
							
							int v = partner.index;
							
							if ( history [ u ] [ v ] > 0 )
							{ }
							else
							{
								history [ u ] [ v ] = 1;
								history [ v ] [ u ] = 1;
								
								return partner;
							}
							
							break;
							
						default :
							
							throw new UnsupportedOperationException ();
					}
				}
			}
        }
        
        return null;
    }
    
    public void go ()
    {
        System.out.println ( String.format ( "go ( \"%s\" )", _social_mechanisms_ ) );
        
		// 
		
		switch ( _history_ )
		{
			case NONE :
				
				break;
				
			case CYCLE :
				
				int n = persons.length;
				
				for ( int u = 0; u < n; u ++ )
				{
					for ( int v = 0; v < n; v ++ )
					{
						history [ u ] [ v ] = 0;
					}
				}
				
				break;
				
			case TOTAL :
				
				break;
				
			default :
				
				throw new UnsupportedOperationException ();
		}
		
		// 
        
        Person [] sequence = shuffle ( persons );
		
        for ( Person a : sequence )
        {
            Person b = one_of ( a );
            
            if ( b != null )
            {
				System.out.println(String.format("%s <> %s",a.name,b.name));
				
                hot_meeting ( a, b );
                
				for ( Link link : a.graph.links ) link.weight_old = link.weight_new;
				for ( Link link : b.graph.links ) link.weight_old = link.weight_new;
				
                hot_person_coh ( a );
                hot_person_coh ( b );
            }
        }
        
        num += 1;
        
        stats ();
    }
	
    public void go ( MediaAgent2 media )
    {
        System.out.println ( String.format ( "go ( \"%s\" %6.4f )", media.source, media.factor ) );
        
        double factor = max( min( media.factor, 1.0 ), 0.0 );
        
        Person [] sequence = shuffle ( persons );
        
        int m = ( int ) ( sequence.length * factor );
        
        for ( int index = 0; index < m; index ++ )
        {
            Person person = sequence [ index ];
                    
			// media agent code
			// 1 media is a dw, use weight
			// 2 media is absolute
			
			Link links [] = new Link [ 8 + 8 + 5 + 5 + 5 * 8 ];
			
			double weights [] = new double [ 8 + 8 + 5 + 5 + 5 * 8 ];
			
			// 
			
			int p;
			
			// 
			
			p = 0;
			
			for ( int j = 0; j < 8; j ++ ) links [ p ++ ] = person.graph.gspalinks [ j ];
			for ( int j = 0; j < 8; j ++ ) links [ p ++ ] = person.graph.gspvlinks [ j ];
			
			for ( int j = 0; j < 5; j ++ ) links [ p ++ ] = person.graph.ispalinks  [ j ];
			for ( int j = 0; j < 5; j ++ ) links [ p ++ ] = person.graph.ispvlinks [ j ];
			
			for ( int u = 0; u < 5; u ++ )
				for ( int v = 0; v < 8; v ++ )
					links [ p ++ ] = person.graph.facilitations [ u ] [ v ];
			
			// 
			
			p = 0;
			
			for ( int j = 0; j < 8; j ++ ) weights [ p ++ ] = media.input_a [ j ];
			for ( int j = 0; j < 8; j ++ ) weights [ p ++ ] = media.input_b [ j ];
			
			for ( int j = 0; j < 5; j ++ ) weights [ p ++ ] = media.input_c [ j ];
			for ( int j = 0; j < 5; j ++ ) weights [ p ++ ] = media.input_d [ j ];
			
			for ( int u = 0; u < 5; u ++ )
				for ( int v = 0; v < 8; v ++ )
					weights [ p ++ ] = media.input_f [ u ] [ v ];
			
			// 
			
			double mw = person.media;
			
			switch ( media.media )
			{
				case RELATIVE :
					
					for ( int j = 0; j < links.length; j ++ )
					{
						if ( nan ( weights [ j ] ) )
						{ }
						else
						{
							Link link = links [ j ];
							
							link.weight_new = link.weight_old + mw * weights [ j ];
						}
					}
					
					break;
					
				case ABSOLUTE :
					
					for ( int j = 0; j < links.length; j ++ )
					{
						if ( nan ( weights [ j ] ) )
						{ }
						else
						{
							Link link = links [ j ];
							
							link.weight_new = weights [ j ];
						}
					}
					
					break;
					
				default :
					
					throw new UnsupportedOperationException ();
			}
			
			// 
			
			for ( Link link : person.graph.links ) link.weight_old = link.weight_new;
			
			// 
			
            hot_person_coh ( person );
        }
        
        num += 1;
        
        stats ();
    }
    
	public boolean nan ( double value )
	{
		return Double.isNaN ( value );
	}
	
    public Person [] shuffle ( Person [] persons )
    {
        int n = persons.length;
        
        Person [] sequence = new Person [ n ];
        
        for ( int index = 0; index < n; index ++ )
        {
            int p = random_int ( n );
            
            while ( sequence [ p % n ] != null ) ++ p;
            
            sequence [ p % n ] = persons [ index ];
        }
        
        return sequence;
    }
    
    public void hot_meeting ( Person a, Person b )
    {
        hot_send ( a.graph, b.graph );
        hot_send ( b.graph, a.graph );
    }
    
    public void hot_send ( Graph sender, Graph receiver )
    {
        if ( false )
        { }
        else if ( _social_mechanisms_.equals ( "contagion" ) )
        {
            contagion  ( sender, receiver );
        }
        else if ( _social_mechanisms_.equals ( "altruism" ) )
        {
            altruism   ( sender, receiver );
        }
        else if ( _social_mechanisms_.equals ( "means-ends" ) )
        {
            means_ends ( sender, receiver );
        }
        else if ( _social_mechanisms_.equals ( "altruism & contagion" ) )
        {
            altruism   ( sender, receiver );
            contagion  ( sender, receiver );
        }
        else if ( _social_mechanisms_.equals ( "means-ends & contagion" ) )
        {
            means_ends ( sender, receiver );
            contagion  ( sender, receiver );
        }
        else if ( _social_mechanisms_.equals ( "means-ends & contagion & altruism" ) )
        {
            means_ends ( sender, receiver );
            contagion  ( sender, receiver );
            altruism   ( sender, receiver );
        }
        else
        {
            throw new UnsupportedOperationException ( _social_mechanisms_ );
        }
    }
    
    public void means_ends ( Graph sender, Graph receiver )
    {
        for ( int a = 0; a < 5; a ++ )
        {
            for ( int b = 0; b < 8; b ++ )
            {
                Link link = sender.facilitations [ a ] [ b ];
                
                if ( ( link.weight_old < - 0.3 ) || ( link.weight_old > + 0.3 ) )
                {
                    Link target = receiver.facilitations [ a ] [ b ];
                    
					target.weight_new = target.weight_old + weighting ( link.weight_old, target.weight_old );
                }
            }
        }
    }
    
	/*
    public double weighting ( double sender, double receiver )
    {
        double distance = sender - receiver;
        
        double weight;
        
        if ( false )
        { }
        else if ( abs ( distance ) > 1.0 ) weight = 0.1;
        else if ( abs ( distance ) > 0.8 ) weight = 0.2;
        else if ( abs ( distance ) > 0.6 ) weight = 0.3;
        else if ( abs ( distance ) > 0.4 ) weight = 0.4;
        else
            weight = 0.5;
        
        return weight * sender;
    }
    */
	
	public boolean _test_compatible_ = false;
	
	public double weighting ( double sender, double receiver )
	{
		if ( _test_compatible_ )
		{
			double distance = sender - receiver;
			
			double weight; 
			
			if ( false )
			{ }
			else if ( abs ( distance ) > 1.0 ) weight = 0.1;
			else if ( abs ( distance ) > 0.8 ) weight = 0.2;
			else if ( abs ( distance ) > 0.6 ) weight = 0.3;
			else if ( abs ( distance ) > 0.4 ) weight = 0.4;
			else
				weight = 0.5;
			
			if ( sender > receiver ) return + abs( weight*sender); else return - abs(weight*sender);
		}
		else
		{
			double diff = 0.0;
			
            if ( abs( sender - receiver ) >= 2.0	) return 0.0;
            	
			// positive Sender
			if ( ( receiver >= 0.6 ) && ( sender > ( + 0.3 ) ) ) diff = + receiver / 100.0 * 8.3;
			if ( ( receiver >= 0.2 ) && ( receiver < 0.6 ) && ( sender > ( + 0.3 ) ) ) diff = + receiver / 100.0 * 7.3; 
			if ( ( receiver >= -0.2 ) && ( receiver < 0.2 ) && ( sender > ( + 0.3 ) ) ) diff = + receiver / 100.0 * 4.0; 
			if ( ( receiver > -0.6 ) && ( receiver < -0.2 ) && ( sender > ( + 0.3 ) ) ) diff = - receiver / 100.0 * 4.1; 
			if ( ( receiver <= -0.6 ) && ( sender > ( + 0.3 ) ) ) diff = - receiver / 100.0 * 3.0; 
			// negative Sender
			if ( ( receiver >= 0.6 ) && ( sender > ( - 0.3 ) ) ) diff = - receiver / 100.0 * 0.3;
			if ( ( receiver >= 0.2 ) && ( receiver < 0.6 ) && ( sender > ( - 0.3 ) ) ) diff = - receiver / 100.0 * 0.6; 
			if ( ( receiver >= -0.2 ) && ( receiver < 0.2 ) && ( sender > ( - 0.3 ) ) ) diff = - receiver / 100.0 * 1.3; 
			if ( ( receiver > -0.6 ) && ( receiver < -0.2 ) && ( sender > ( - 0.3 ) ) ) diff = - receiver / 100.0 * 0.3; 
			if ( ( receiver <= -0.6 ) && ( sender > ( - 0.3 ) ) ) diff = - receiver / 100.0 * 2.2; 
			
			
			return diff;
		}
	}
	
	
    public void contagion ( Graph sender, Graph receiver )
    {
        for ( int index = 0; index < 5; index ++ )
        {
            Unit issue = sender.issues [ index ];
            
            if ( abs ( issue.valence ) > _valence_threshold_ )
            {
                Link target = receiver.ispvlinks [ index ]; 
                
                Unit unit_a =   sender.issues [ index ];
                Unit unit_b = receiver.issues [ index ];
                
                target.weight_new = target.weight_old + weighting_contagion ( unit_a.valence, unit_b.valence );
            }
        }
    }

	
	public double weighting_contagion ( double sender, double receiver )
	{
		double diff = 0.0;
				
		// positive Sender
		if ( ( receiver >= 0.6 ) && ( sender > ( + 0.3 ) ) ) diff = + receiver / 100.0 * 7.5; // probable to high, because of the stronger effects of emotional contagion on the decision results the empirical values are halved 
		if ( ( receiver >= 0.2 ) && ( receiver < 0.6 ) && ( sender > ( + 0.3 ) ) ) diff = + receiver / 100.0 * 3.50; 
		if ( ( receiver >= -0.2 ) && ( receiver < 0.2 ) && ( sender > ( + 0.3 ) ) ) diff = + receiver / 100.0 * 0.6; 
		if ( ( receiver > -0.6 ) && ( receiver < -0.2 ) && ( sender > ( + 0.3 ) ) ) diff = - receiver / 100.0 * 1.0; 
		if ( ( receiver <= -0.6 ) && ( sender > ( + 0.3 ) ) ) diff = - receiver / 100.0 * 2.5; 
		// negative Sender
		if ( ( receiver >= 0.6 ) && ( sender > ( - 0.3 ) ) ) diff = + receiver / 100.0 * 4.0;
		if ( ( receiver >= 0.2 ) && ( receiver < 0.6 ) && ( sender > ( - 0.3 ) ) ) diff = + receiver / 100.0 * 0.35; 
		if ( ( receiver >= -0.2 ) && ( receiver < 0.2 ) && ( sender > ( - 0.3 ) ) ) diff = - receiver / 100.0 * 0.10; 
		if ( ( receiver > -0.6 ) && ( receiver < -0.2 ) && ( sender > ( - 0.3 ) ) ) diff = - receiver / 100.0 * 0.85; 
		if ( ( receiver <= -0.6 ) && ( sender > ( - 0.3 ) ) ) diff = - receiver / 100.0 * 1.8; 
		
		return diff;
	}
	
    public void altruism ( Graph sender, Graph receiver )
    {
        throw new UnsupportedOperationException ( "altruism" );
    }
    
    public double random_double ()
    {
        return random.nextDouble ();
    }
    
    public int random_int ( int sup )
    {
        return random.nextInt ( sup );
    }
    
    public int round ( double number )
    {
        return ( ( int ) number );
    }
    
    public void stats_header ()
    {
        try
        {
            BufferedWriter io = new BufferedWriter ( new FileWriter ( file, true ) );
            
            try
            {
                io.write ( String.format( "#%s\t%s\t", "Subject", "Tick" ) );
                
                for ( String name : issues )
                {
                    io.write ( String.format ( "%s-Activation\t%s-Valence\t", name, name ) );
                }
                
                for ( String name : goals )
                {
                    io.write ( String.format ( "%s-Activation\t%s-Valence\t", name, name ) );
                }
                
                io.write ( String.format( "%s\r\n", "Cluster" ) );
            }
            finally
            {
                io.close ();
            }
        }
        catch ( Exception exception )
        {
            throw new UnsupportedOperationException ( exception );
        }
    }
    
    public void stats ()
    {
        try
        {
            BufferedWriter io = new BufferedWriter ( new FileWriter ( file, true ) );
            
            try
            {
                for ( Person person : persons )
                {
                    io.write ( String.format( Locale.ENGLISH, "%12s\t%6d\t", person.name, num ) );
                    
                    for ( Unit unit : person.graph.issues )
                    {
                        io.write ( String.format ( Locale.ENGLISH, "%+6.4f\t%+6.4f\t", unit.activation, unit.valence ) );
                    }
                    
                    for ( Unit unit : person.graph.goals )
                    {
                        io.write ( String.format ( Locale.ENGLISH, "%+6.4f\t%+6.4f\t", unit.activation, unit.valence ) );
                    }
                    
                    io.write ( String.format ( Locale.ENGLISH, "%6d\n", person.cluster ) );
                }
            }
            finally
            {
                io.close ();
            }
        }
        catch ( Exception exception )
        {
            throw new UnsupportedOperationException ( exception );
        }
    }
    
    public static String format ( String file )

    {
        SimpleDateFormat str = new SimpleDateFormat ( "yyMMddHHmmss" );
        return String.format ( "XXX.txt", file, str.format ( new Date () ) );
				
    }
    
   public static void main ( String [] args )

    {
        // start a new simulation with a log file called output ( date/time added )       
        for ( int indexN = 0; indexN < 1	; indexN ++ )
        {
			
            HOTCOInnoSimV15 simulation = new HOTCOInnoSimV15 ( format ( "output" ) ); // new HOTCOInnoSimV15 ( format ( "output" ), 1234 ); // Random Seed = 1234

         // set parameter        

            simulation._network_type_                   = "social-radius"; // "social-radius";
            simulation._social_mechanisms_              = "means-ends & contagion"; // "means-ends & contagion";       

            simulation._goal_excit_                     = + 0.04;
            simulation._excit_weight_                   = + 0.04;
            simulation._inhib_weight_                   = + 0.04; // + 0.06
            simulation._max_activation_                 = + 0.99;
            simulation._min_activation_                 = - 0.99;
            simulation._decay_amount_                   = + 0.05;
            simulation._output_threshold_               = - 0.99;
            simulation._valence_weight_                 = + 0.05;
            simulation._random_partners_                = + 0.05;
            simulation._ex_radius_                      = + 0.001; 
            simulation._link_chance_                    = + 0.50;
            simulation._activation_default_value_       = + 0.01;
            simulation._new_activation_default_value_   = NaN;
            simulation._valence_default_value_          = + 0.01;
            simulation._new_valence_default_value_      = + 0.01;
            simulation._valence_threshold_              = + 0.10;
            simulation._activated_                      = + 0.20;
        

            simulation._asymptote_                      = + 0.0001;

        

		simulation._history_ = HISTORY.CYCLE;	
		simulation._test_compatible_ = false;

		// HISTORY.NONE;
		// HISTORY.CYCLE;
		// HISTORY.TOTAL;
        // setup
        
			simulation.setup ( "InputFile.txt" );
            
		print_units ( simulation );
		simulation.go ();
		print_units ( simulation );

		// apply some media agent with a custom factor
			
			simulation.go ( new MediaAgent2 ( "MediaAgent.txt", MEDIA.RELATIVE, 0.7 ) );				
           
        for ( int index = 0; index < 8; index ++ )

        {
            simulation.go ();
        }

        // apply some other media agent with a custom factor 
			
			simulation.go ( new MediaAgent2 ( "MediaAgent2.txt", MEDIA.RELATIVE, 0.7 ) );
			
           
        for ( int index = 0; index < 9; index ++ )
			
        {
            simulation.go ();
        }
		
        								
	  System.out.println( String.format( "Results written to \"%s\"", simulation.file ) );
		
    
     }
 }	

	public static void print_units ( HOTCOInnoSimV15 simulation )
	{

		for ( Person person : simulation.persons )

		{

			System.out.print( String.format( "%s\t", person.name ) );

			

			for ( int index = 0; index < 5; index ++ )

			{

				System.out.print( String.format( "%+6.4f\t", person.graph.issues[index].activation ) );

			}		

			System.out.println();

		}

	}

}

